from .. import settings
from .common import validate_package_config, load_config


BOILERPLATE_REQUIRED_CONFIG_KEYS = (
    'package-name',
    'identifier',
    'version',
    'templates',
)


def validate_boilerplate_config(config, path):
    errors = []

    errors += validate_package_config(
        config=config,
        required_keys=BOILERPLATE_REQUIRED_CONFIG_KEYS,
        config_path=path
    )

    templates = config.get('templates', [])
    if isinstance(templates, (list, tuple)):
        for template in templates:
            if not isinstance(template, (list, tuple)) or len(template) != 2:
                errors.append(
                    'Templates must be a list/tuple of lists/tuples '
                    'with two items each.'
                )

    excluded_files = config.get('excluded', [])
    if not isinstance(excluded_files, (list, tuple)):
        errors.append('Included files setting must be a list or a tuple.')

    if config.get('protected'):
        errors.append(
            'The `protected` section is deprecated and not used '
            'anymore. By default, all files will be included in '
            'the boilerplate. If you wish to exclude some files, '
            'use the `excluded` section.'
        )

    return errors


def validate_boilerplate(path=None):
    config = load_config(settings.BOILERPLATE_CONFIG_FILENAME, path)
    return validate_boilerplate_config(config, path)



import shutil
import imp
import time
import os

import click

from .. import settings
from .. import messages
from ..utils import silence_stderr, create_temp_dir
from .common import validate_package_config, load_config


ADDON_REQUIRED_CONFIG_KEYS = (
    'package-name',
)


def validate_aldryn_config_py(path):
    aldryn_config_path = os.path.join(path, 'aldryn_config.py')
    if os.path.exists(aldryn_config_path):
        temp_dir = create_temp_dir()
        try:
            shutil.copy(aldryn_config_path, temp_dir)
            temp_path = os.path.join(temp_dir, 'aldryn_config.py')
            try:
                with silence_stderr():
                    # suppressing RuntimeWarning: Parent module 'aldryn_config'
                    # not found while handling absolute import

                    # randomizing source name
                    source = 'aldryn_config.config_{}'.format(int(time.time()))
                    module = imp.load_source(source, temp_path)

                # checking basic functionality of the Form
                form = module.Form({})
                form.is_valid()

            except Exception:
                # intentionally catch every exception
                import traceback
                click.secho(
                    "An error occurred during validating 'aldryn_config.py'. "
                    "Please check the exception below:\n",
                    fg='red'
                )
                raise click.ClickException(traceback.format_exc())
        finally:
            shutil.rmtree(temp_dir)


def validate_addon(path=None):
    setup_py_path = os.path.join(path or '.', 'setup.py')
    if not os.path.exists(setup_py_path):
        raise click.ClickException(
            messages.FILE_NOT_FOUND.format(setup_py_path)
        )

    config = load_config(settings.ADDON_CONFIG_FILENAME, path)
    validate_aldryn_config_py(path)
    validate_package_config(config, ADDON_REQUIRED_CONFIG_KEYS, path)



import json
import os

from .. import exceptions
from .. import messages


VALID_LICENSE_FILENAMES = (
    'LICENSE.txt',
    'LICENSE',
    'license.txt',
    'license',
)


def get_license(path):
    for fname in VALID_LICENSE_FILENAMES:
        fpath = os.path.join(path, fname)
        if os.path.exists(fpath):
            return fpath


def load_config(fname, path=None):
    config_fpath = os.path.join(path or '.', fname)
    if not os.path.exists(config_fpath):
        raise exceptions.AldrynException(
            messages.FILE_NOT_FOUND.format(config_fpath)
        )

    with open(config_fpath) as f:
        try:
            return json.load(f)
        except ValueError:
            raise exceptions.AldrynException(
                'Config file could not be loaded: Invalid JSON'
            )


def validate_package_config(config, required_keys, config_path):
    errors = []

    if not get_license(config_path):
        errors.append(
            'Required LICENSE file not found. Valid names are {}.'.format(
                ', '.join(VALID_LICENSE_FILENAMES)
            )
        )

    for key in required_keys:
        if key not in config:
            errors.append(
                'Required key "{}" not found in config file.'.format(key)
            )

    return errors



import os
import subprocess
import tarfile

import click
from io import BytesIO

from .. import settings
from ..utils import (
    dev_null, tar_add_bytesio, get_package_version, create_temp_dir,
    tar_add_stringio, get_bytes_io, get_string_io,
)
from ..validators.addon import validate_addon

from .common import add_meta_files


def package_addon(path):
    temp_dir = create_temp_dir()
    with dev_null() as devnull:
        subprocess.check_call(
            ['python', 'setup.py', 'sdist', '-d', temp_dir],
            cwd=path, stdout=devnull,
        )
    for filename in os.listdir(temp_dir):
        if filename.endswith('.tar.gz'):
            return os.path.join(temp_dir, filename)

    raise click.ClickException('Packaged addon could not be found')


def add_addon_meta_files(tar, path):
    # aldryn_config.py
    try:
        with open(os.path.join(path, 'aldryn_config.py')) as fobj:
            tar_add_bytesio(tar, BytesIO(fobj.read()), 'aldryn_config.py')
    except (OSError, IOError):
        click.secho(
            'Warning: Aldryn config file \'aldryn_config.py\' not found. '
            'Your app will not have any configurable settings.',
            fg='yellow'
        )

    # version
    tar_add_stringio(tar, get_string_io(get_package_version(path)), 'VERSION')


def create_addon_archive(path):
    data = get_bytes_io()

    with tarfile.open(mode='w:gz', fileobj=data) as tar:
        add_meta_files(tar, path, settings.ADDON_CONFIG_FILENAME)
        add_addon_meta_files(tar, path)
        packaged_addon = package_addon(path)
        tar.add(packaged_addon, arcname='package.tar.gz')

    data.seek(0)
    return data


def upload_addon(client, path=None):
    path = path or '.'
    validate_addon(path)
    archive_obj = create_addon_archive(path)
    return client.upload_addon(archive_obj)



import json
import sys
import os
from time import time

import click

from ..utils import check_output, is_windows, check_call
from .. import exceptions
from .. import settings


def get_aldryn_project_settings(path=None):
    project_home = get_project_home(path)
    with open(os.path.join(project_home, settings.ALDRYN_DOT_FILE)) as fh:
        return json.load(fh)


def get_project_home(path=None):
    """
    find project root by traversing up the tree looking for
    the '.aldryn' file
    """
    previous_path = None
    current_path = path or os.getcwd()

    # loop until we're at the root of the volume
    while current_path != previous_path:

        # check if '.aldryn' file exists in current directory
        dotfile = os.path.join(current_path, settings.ALDRYN_DOT_FILE)
        if os.path.exists(dotfile):
            return current_path

        # traversing up the tree
        previous_path = current_path
        current_path = os.path.abspath(os.path.join(current_path, os.pardir))

    raise click.ClickException(
        "Aldryn project file '.aldryn' could not be found! Please make sure "
        "you're in an Aldryn project folder and the file exists."
    )


UNIX_DOCKER_COMPOSE_FILENAME = 'docker-compose.yml'
WINDOWS_DOCKER_COMPOSE_FILENAME = 'docker-compose-windows.yml'


def get_docker_compose_cmd(path):
    if is_windows():
        docker_compose_filename = WINDOWS_DOCKER_COMPOSE_FILENAME
        ensure_windows_docker_compose_file_exists(path)
    else:
        docker_compose_filename = UNIX_DOCKER_COMPOSE_FILENAME

    docker_compose_base = [
        'docker-compose', '-f', os.path.join(path, docker_compose_filename)
    ]

    def docker_compose(*commands):
        return docker_compose_base + [cmd for cmd in commands]

    return docker_compose


def ensure_windows_docker_compose_file_exists(path):
    """
    Unfortunately, docker-compose is not yet officially released
     for Windows There's still some rough edges, and volume
     configuration is one. There's also some open issues in
     boot2docker for windows which makes things difficult.

    We have to change the volume specifications to a very specific
     format:

     - absolute paths: relative one's are not yet supported
     - currently only works if the project is running on the C:\ drive
     - unix style paths: need to replace '\' with '/'
     - paths have to start with /c/ instead of C:\ otherwise
        docker-compose gets confused because they use : as separation

    Example:
      unix format:  .:/app:rw
      cwd:          C:\\Users\\aldryn\\acme-portfolio
      windows:     /c/Users/aldryn/acme-portfolio:/app:rw

    Hope that's all. And of course, I'm sorry.
    """

    import yaml

    windows_path = os.path.join(path, WINDOWS_DOCKER_COMPOSE_FILENAME)
    if os.path.isfile(windows_path):
        return

    unix_path = os.path.join(path, UNIX_DOCKER_COMPOSE_FILENAME)
    if not os.path.isfile(unix_path):
        # TODO: use correct exit from click
        click.secho(
            'docker-compose.yml not found at {}'.format(unix_path),
            fg='red',
        )
        sys.exit(1)

    with open(unix_path, 'r') as fh:
        config = yaml.load(fh)

    for component, sections in config.items():
        if 'volumes' not in sections:
            continue
        volumes = []
        for volume in sections['volumes']:
            parts = volume.split(':')
            if len(parts) == 2:
                old_host, container = parts
                mode = None
            else:
                old_host, container, mode = parts

            # assuming relative path's for old_host
            new_host = os.path.abspath(os.path.join(path, old_host))
            # replace C:\ with /c/, because, docker on windows
            new_host = new_host.replace('C:\\', '/c/')
            # change to unix paths
            new_host = new_host.replace('\\', '/')
            new_volume = [new_host, container]
            if mode:
                new_volume.append(mode)
            volumes.append(':'.join(new_volume))

        config[component]['volumes'] = volumes

    with open(windows_path, 'w+') as fh:
        yaml.safe_dump(config, fh)


def get_db_container_id(path, raise_on_missing=True):
    docker_compose = get_docker_compose_cmd(path)
    output = check_output(docker_compose('ps', '-q', 'db')).rstrip(os.linesep)
    if not output and raise_on_missing:
        raise exceptions.AldrynException('Unable to find database container')
    return output


def start_database_server(docker_compose):
    start_db = time()
    click.secho(' ---> Starting local database server')
    click.secho('      ', nl=False)
    check_call(docker_compose('up', '-d', 'db'))
    click.secho('      [{}s]'.format(int(time() - start_db)))



import json
import time
import os
from distutils.version import StrictVersion

import click

from . import utils, __version__


class Config(object):
    config_name = '.aldryn'

    def __init__(self):
        super(Config, self).__init__()
        home = os.path.expanduser('~')
        self.config_path = os.path.join(home, self.config_name)
        self.read()

    def read(self):
        try:
            with open(self.config_path, 'r') as fh:
                config = json.load(fh)
        except IOError:
            # file doesn't exist
            config = {}
        except ValueError:
            # invalid config
            config = {}
        self.config = config

    def save(self):
        with open(self.config_path, 'w+') as fh:
            json.dump(self.config, fh)

    def check_for_updates(self):
        """check daily for updates"""
        if self.config.get('disable_update_check', False):
            return

        timestamp_key = 'update_check_timestamp'
        version_key = 'update_check_version'

        last_checked = self.config.get(timestamp_key, None)
        now = int(time.time())
        current_version = StrictVersion(__version__)

        if not last_checked or last_checked < now - (60 * 60 * 24):
            # try to access PyPI to get the latest available version
            newest_version, _ = utils.get_latest_version_from_pypi()

            if newest_version:
                if newest_version > current_version:
                    self.config[version_key] = str(newest_version)
                self.config[timestamp_key] = now
                self.save()
            elif newest_version is False:
                # fail silently, nothing the user can do about this
                self.config.pop(version_key, None)

        newer_version_string = self.config.get(version_key, None)
        if newer_version_string:
            newer_version = StrictVersion(newer_version_string)
            if newer_version <= current_version:
                self.config.pop(version_key)
                self.save()
            else:
                click.secho(
                    "New version ({version}) available on PyPI. Update "
                    "now using 'pip install aldryn-client=={version}'"
                    .format(version=newer_version),
                    fg='yellow'
                )

    def skip_doctor(self):
        return self.config.get('skip_doctor')

    def get_skip_doctor_checks(self):
        checks = self.config.get('skip_doctor_checks')
        if not checks or not isinstance(checks, list):
            return []
        return checks



import json

from ..utils import tar_add_stringio, get_string_io
from ..validators.common import load_config, get_license


def add_meta_files(tar, path, config_filename):
    # config json file
    config_json = load_config(config_filename, path)
    config_json_fobj = get_string_io()
    json.dump(config_json, config_json_fobj)
    tar_add_stringio(tar, config_json_fobj, config_filename)

    # license
    license_filepath = get_license(path)
    if license_filepath:
        tar.add(license_filepath, 'LICENSE.txt')



import os

from six import with_metaclass


VALID_FIELD_TYPES = [
    'text',
    'checkbox',
    'number',
    'select',
    'x-static-file',
]


class ValidationError(Exception):
    def __init__(self, message):
        self.message = message


class BaseField(object):
    field_type = None
    _order_counter = 0

    def __init__(self, label, required=True, help_text=None, initial=None):
        self.label = label
        self.required = required
        self.help_text = help_text
        self.initial = initial
        self._order_counter = BaseField._order_counter
        BaseField._order_counter += 1

    def clean(self, value):
        return value

    def serialize(self):
        return {
            'field_type': self.field_type,
            'label': self.label,
            'required': self.required,
            'help_text': self.help_text,
            'initial': self.initial,
        }


class CharField(BaseField):
    field_type = 'text'

    def __init__(self, label, min_length=None, max_length=None, required=True,
                 **kwargs):
        super(CharField, self).__init__(label, required, **kwargs)
        self.min_length = min_length
        self.max_length = max_length

    def clean(self, value):
        if not value:
            value = ''
        length = len(value)
        if self.min_length and length < self.min_length:
            raise ValidationError(
                'Value must be at least {} characters, got {} characters'
                .format(self.min_length, length)
            )
        if self.max_length and length > self.max_length:
            raise ValidationError(
                'Value must be less than {} characters, got {} characters'
                .format(self.max_length, length)
            )
        return value

    def serialize(self):
        data = super(CharField, self).serialize()
        data['min_length'] = self.min_length
        data['max_length'] = self.max_length
        return data


class CheckboxField(BaseField):
    field_type = 'checkbox'

    def clean(self, value):
        return bool(value)


class SelectField(BaseField):
    field_type = 'select'

    def __init__(self, label, choices, required=True, **kwargs):
        super(SelectField, self).__init__(label, required, **kwargs)
        self.choices = choices

    def serialize(self):
        data = super(SelectField, self).serialize()
        data['choices'] = self.choices
        return data


class NumberField(BaseField):
    field_type = 'number'

    def __init__(self, label, min_value=None, max_value=None, required=True,
                 **kwargs):
        super(NumberField, self).__init__(label, required, **kwargs)
        self.min_value = min_value
        self.max_value = max_value

    def clean(self, value):
        if not value:
            return value
        if not value.isdigit():
            raise ValidationError('Expected number, but got {}'.format(value))
        value = int(value)
        if self.min_value is not None and value < self.min_value:
            raise ValidationError(
                'Must be bigger than{}'.format(self.min_value)
            )
        if self.max_value is not None and value > self.max_value:
            raise ValidationError(
                'Must be smaller than {}'.format(self.max_value)
            )
        return value


class StaticFileField(BaseField):
    field_type = 'x-static-file'

    def __init__(self, label, extensions=None, required=True, **kwargs):
        super(StaticFileField, self).__init__(label, required, **kwargs)
        self.extensions = extensions

    def clean(self, value):
        if not value:
            return value
        extension = os.path.splitext(value)[1][1:]
        if self.extensions is not None and extension not in self.extensions:
            raise ValidationError(
                'Please choose a file with one of the following extensions: '
                .format(', '.join(self.extensions))
            )
        return value

    def serialize(self):
        data = super(StaticFileField, self).serialize()
        data['extensions'] = self.extensions
        return data


class FormMeta(type):
    def __new__(cls, name, bases, attrs):
        fields = []
        for key, value in attrs.items():
            if isinstance(value, BaseField):
                fields.append((key, value))
        # restore the fields' order as it was in the Form's class body
        attrs['_fields'] = sorted(fields, key=lambda kv: kv[1]._order_counter)
        return super(FormMeta, cls).__new__(cls, name, bases, attrs)


class BaseForm(with_metaclass(FormMeta, object)):
    def __init__(self, data=None):
        self.data = data or {}

    def serialize(self):
        form = []
        for name, field in self._fields:
            form.append((name, field.serialize()))
        return form

    def is_valid(self):
        self.cleaned_data = {}
        self.errors = {}
        try:
            self.clean()
        except ValidationError as e:
            self.errors[None] = e.message
        return not self.errors

    def clean(self):
        for name, field in self._fields:
            value = self.data.get(name, None)
            if field.required and value is None:
                self.errors[name] = 'This field is required'
            else:
                try:
                    self.cleaned_data[name] = field.clean(value)
                except ValidationError as e:
                    self.errors[name] = e.message
        return self.cleaned_data

    def to_settings(self, data, settings):
        return settings

    def save(self):
        return self.cleaned_data



# -*- coding: utf-8 -*-
import os
import subprocess
import sys
from collections import OrderedDict

import click

from . import cloud
from . import utils

ERROR = 1
WARNING = 0


class Check(object):
    name = None
    command = None
    error_level = ERROR

    def run_check(self):
        errors = []
        try:
            utils.check_call(self.command, catch=False, silent=True)
        except OSError as exc:
            if exc.errno == os.errno.ENOENT:
                errors.append('executable {} not found'.format(self.command[0]))
            else:
                msg = (
                    "Command '{}' returned non-zero exit status {}"
                    .format(self.fmt_command(), exc.errno)
                )
                if hasattr(exc, 'strerror'):
                    msg += ': {}'.format(exc.strerror)

                errors.append(msg)
        except subprocess.CalledProcessError as exc:
            errors += self.fmt_exception(exc)
        return errors

    def fmt_command(self):
        return ' '.join(self.command)

    def fmt_exception(self, exc):
        command_output = exc.output

        if command_output:
            message = command_output
        else:
            message = (
                "Command '{}' returned non-zero exit status {}"
                .format(self.fmt_command(), exc.returncode)
            )

        return [message]


class LoginCheck(Check):
    name = 'Login'

    def run_check(self):
        client = cloud.CloudClient(cloud.get_endpoint())
        success, msg = client.check_login_status()
        if not success:
            return [msg]


class GitCheck(Check):
    name = 'Git'
    command = ('git', '--version')


class DockerClientCheck(Check):
    name = 'Docker Client'
    command = ('docker', '--version')


class DockerMachineCheck(Check):
    name = 'Docker Machine'
    command = ('docker-machine', '--version')
    error_level = WARNING


class DockerComposeCheck(Check):
    name = 'Docker Compose'
    command = ('docker-compose', '--version')


def get_engine_down_error():
    msg = "Couldn't connect to Docker daemon. "
    if utils.is_linux():
        msg += 'Please start the docker service.'
    else:
        msg += 'You might need to run `docker-machine start`.'
    return msg


class DockerEngineBaseCheck(Check):
    def fmt_exception(self, exc):
        errors = super(DockerEngineBaseCheck, self).fmt_exception(exc)
        if exc.returncode == 125:
            errors.append(get_engine_down_error())
        return errors


class DockerEngineCheck(DockerEngineBaseCheck):
    name = 'Docker Engine Connectivity'
    command = ('docker', 'run', '--rm', 'busybox', 'true')

    def fmt_exception(self, exc):
        errors = super(DockerEngineCheck, self).fmt_exception(exc)
        if not utils.is_windows():
            default_host_path = '/var/run/docker.sock'
            default_host_url = 'unix://{}'.format(default_host_path)
            current_host_url = os.environ.get('DOCKER_HOST')
            current_host_is_default = current_host_url == default_host_url

            # run additional checks if it user is running default config
            if not current_host_url or current_host_is_default:

                # check if docker socket exists
                if not os.path.exists(default_host_path):
                    errors.append(
                        'Could not find docker engine socket at {}. Please '
                        'make sure your docker engine is setup correctly and '
                        'check the docker installation guide: '
                        'https://docs.docker.com/engine/installation/'
                        .format(default_host_path)
                    )

                elif not os.access(default_host_path, os.R_OK):
                    # check if docker socket is readable
                    errors.append(
                        'No read permissions on {}. Please make sure the unix '
                        'socket can be accessed without root permissions. '
                        'More information can be found in the docker '
                        'installation guide: https://docs.docker.com/engine/'
                        'installation/linux/ubuntulinux/#create-a-docker-group'
                        .format(default_host_path)
                    )

        return errors


class DockerEnginePingCheck(DockerEngineBaseCheck):
    name = 'Docker Engine Internet Connectivity'
    command = (
        'docker', 'run', '--rm', 'busybox', 'ping',
        '-c', '1',  # stop after one packet response
        '-W', '5',  # timeout of 5 seconds
        '8.8.8.8',
    )

    def fmt_exception(self, exc):
        errors = super(DockerEnginePingCheck, self).fmt_exception(exc)
        errors.append(
            "The 'ping' command inside docker is not able to ping "
            '8.8.8.8. This might be due to missing internet connectivity, '
            'a firewall or a network configuration problem.'
        )
        return errors


class DockerEngineDNSCheck(DockerEngineBaseCheck):
    name = 'Docker Engine DNS Connectivity'
    command = (
        'docker', 'run', '--rm', 'busybox',
        'sh', '-c',  # run in new a shell to avoid problems with timeout
        'timeout -t 5 nslookup aldryn.com',
    )

    def fmt_exception(self, exc):
        errors = super(DockerEngineDNSCheck, self).fmt_exception(exc)
        errors.append(
            'The DNS resolution inside docker is not able to resolve '
            'aldryn.com. This might be due to missing internet connectivity, '
            'a firewall or a network configuration problem.'
        )
        return errors


ALL_CHECKS = OrderedDict([
    ('login', LoginCheck),
    ('git', GitCheck),
    ('docker-client', DockerClientCheck),
    ('docker-machine', DockerMachineCheck),
    ('docker-compose', DockerComposeCheck),
    ('docker-server', DockerEngineCheck),
    ('docker-server-ping', DockerEnginePingCheck),
    ('docker-server-dns', DockerEngineDNSCheck),
])


def check_requirements(config=None, checks=None):
    if checks is None:
        checks = ALL_CHECKS.keys()

    skip_doctor_checks = config.get_skip_doctor_checks() if config else []

    for check_key in checks:
        if check_key in skip_doctor_checks:
            continue
        check = ALL_CHECKS.get(check_key)
        if not check:
            click.secho(
                'Invalid check {}'.format(check_key),
                fg='red'
            )
            sys.exit(1)
        errors = check().run_check()
        yield check_key, check.name, errors


def get_prefix(success):
    is_windows = utils.is_windows()
    if success:
        symbol = ' OK  ' if is_windows else ' ✓  '
        color = 'green'
    else:
        symbol = ' ERR ' if is_windows else ' ✖  '
        color = 'red'
    return symbol, color


def check_requirements_human(config, checks=None, silent=False):
    errors = []

    if config and config.skip_doctor():
        return True

    for check, check_name, error in check_requirements(config, checks):
        if error:
            errors.append((check, check_name, error))
        if not silent:
            symbol, color = get_prefix(not error)
            click.secho(symbol, fg=color, nl=False)
            click.secho(check_name)

    if not errors:
        return True

    if not silent:
        click.secho('\nThe following errors occurred:', fg='red')
        for check, check_name, msgs in errors:
            click.secho('\n {}:'.format(check_name))
            for msg in msgs:
                click.secho(' > {}'.format(msg))

    max_error_level = max([ALL_CHECKS[e[0]].error_level for e in errors])
    return True if max_error_level == 0 else False



import json
import tarfile
import re
import os
import stat
import subprocess
import sys
from time import sleep, time

import click
import requests
import shutil

from ..utils import (
    check_call, check_output, is_windows, pretty_size, get_size,
    download_file,
)
from ..cloud import get_aldryn_host
from .. import settings
from . import utils


DEFAULT_GIT_HOST = 'git@git.{aldryn_host}'
GIT_CLONE_URL = '{git_host}:{project_slug}.git'


def get_git_host():
    git_host = os.environ.get('ALDRYN_GIT_HOST')
    if git_host:
        click.secho('Using custom git host {}\n'.format(git_host), fg='yellow')
    else:
        git_host = DEFAULT_GIT_HOST.format(aldryn_host=get_aldryn_host())
    return git_host


def get_git_clone_url(slug):
    return GIT_CLONE_URL.format(git_host=get_git_host(), project_slug=slug)


def clone_project(website_slug, path):
    click.secho('\ncloning project repository', fg='green')
    website_git_url = get_git_clone_url(website_slug)
    clone_args = ['git', 'clone', website_git_url]
    if path:
        clone_args.append(path)

    check_call(clone_args)


def configure_project(website_slug, path, client):
    website_id = client.get_website_id_for_slug(website_slug)

    # Detect old style or invalid projects
    compose_config = os.path.join(path, 'docker-compose.yml')
    if not os.path.isfile(compose_config):
        raise click.ClickException(
            "Could not find a valid 'docker-compose.yml' file. Please make "
            "sure that this project has been updated on Aldryn to use "
            "Base Project version 3 or higher."
        )

    # create .aldryn file
    website_data = {'id': website_id, 'slug': website_slug}
    with open(os.path.join(path, settings.ALDRYN_DOT_FILE), 'w+') as fh:
        json.dump(website_data, fh)


def setup_website_containers(client, stage, path):
    docker_compose = utils.get_docker_compose_cmd(path)

    existing_db_container_id = utils.get_db_container_id(path, False)

    # stop all running for project
    check_call(docker_compose('stop'))

    # pull docker images
    click.secho('downloading remote docker images', fg='green')
    check_call(docker_compose('pull'))

    # build docker images
    click.secho('building local docker images', fg='green')
    check_call(docker_compose('build'))

    if existing_db_container_id:
        click.secho('removing old database container', fg='green')
        check_call(docker_compose('stop', 'db'))
        check_call(docker_compose('rm', '-f', 'db'))

    click.secho('creating new database container', fg='green')
    ImportRemoteDatabase(client=client, stage=stage, path=path)()

    click.secho('sync and migrate database', fg='green')

    if is_windows():
        # interactive mode is not yet supported with docker-compose
        # on windows. that's why we have to call it as daemon
        # and just wait a sane time
        check_call(docker_compose('run', '-d', 'web', 'start', 'migrate'))
        sleep(30)
    else:
        check_call(docker_compose('run', 'web', 'start', 'migrate'))


def create_workspace(client, website_slug, stage, path=None, force_overwrite=False):
    click.secho('Creating workspace', fg='green')

    path = os.path.abspath(
        os.path.join(path, website_slug)
        if path else website_slug
    )

    if os.path.exists(path) and (not os.path.isdir(path) or os.listdir(path)):
        if force_overwrite or click.confirm(
            'The path {} already exists and is not an empty directory. '
            'Do you want to remove it and continue?'.format(path)
        ):
            if os.path.isdir(path):
                def del_rw(action, name, exc):
                    os.chmod(name, stat.S_IWRITE)
                    os.remove(name)
                shutil.rmtree(path, onerror=del_rw)
            else:
                os.remove(path)
        else:
            click.secho('Aborting', fg='red')
            sys.exit(1)

    # clone git project
    clone_project(website_slug, path)

    # check for new baseproject + add .aldryn
    configure_project(website_slug, path, client)

    # setup docker website containers
    setup_website_containers(client, stage, path)

    # download media files
    pull_media(client, stage, path)

    instructions = (
        "Finished setting up your project's workspace!",
        "To start the project, please:",
        " - change directory into '{}'".format(path),
        " - run 'aldryn project up'",
    )

    click.secho('\n\n{}'.format(os.linesep.join(instructions)), fg='green')


class DatabaseImportBase(object):
    database_extensions = ['hstore', 'postgis']
    restore_commands = {
        'sql': 'psql -U postgres db < {}',
        'binary': (
            'pg_restore -U postgres --dbname=db -n public '
            '--no-owner --exit-on-error {}'
        ),
        'archived-binary': (
            'tar -xzOf {}'
            ' | pg_restore -U postgres --dbname=db -n public '
            '--no-owner --exit-on-error'
        ),
    }

    def __init__(self, *args, **kwargs):
        super(DatabaseImportBase, self).__init__()
        self.client = kwargs.pop('client')
        self.path = kwargs.pop('path', None) or utils.get_project_home()
        self.website_id = utils.get_aldryn_project_settings(self.path)['id']
        self.website_slug = utils.get_aldryn_project_settings(self.path)['slug']
        self.docker_compose = utils.get_docker_compose_cmd(self.path)

        self.start_time = time()

    def __call__(self, *args, **kwargs):
        return self.run()

    def setup(self):
        raise NotImplementedError

    def run(self):
        self.setup()
        self.prepare_db_server()
        self.restore_db()
        self.finish()

    def prepare_db_server(self):
        utils.start_database_server(self.docker_compose)

        click.secho(' ---> Waiting for local database server', nl=False)

        db_container_id = utils.get_db_container_id(self.path)

        start_wait = time()
        # check for postgres in db container to start

        # sleep 5 seconds initially because the server is quickly
        # available during startup, but will go down again to
        # create the initial database. We're giving postgres a head start
        sleep(5)

        for attempt in range(10):
            try:
                check_call([
                    'docker', 'exec', db_container_id,
                    'ls', '/var/run/postgresql/.s.PGSQL.5432',
                ], catch=False, silent=True)
            except subprocess.CalledProcessError:
                sleep(5)
            else:
                break
        else:
            click.secho(
                "Couldn't connect to database container. "
                "Database server may not have started.",
                fg='red',
            )
            sys.exit(1)
        click.echo(' [{}s]'.format(int(time() - start_wait)))

        # drop any existing connections
        check_call([
            'docker', 'exec', db_container_id,
            'psql', '-U', 'postgres', '-c',
            "SELECT pg_terminate_backend(pg_stat_activity.pid) "
            "FROM   pg_stat_activity "
            "WHERE  pg_stat_activity.datname = 'db' "
            "  AND  pid <> pg_backend_pid();",
        ], silent=True)

        click.secho(' ---> Removing local database', nl=False)
        start_remove = time()
        # create empty db
        subprocess.call([
            'docker', 'exec', db_container_id,
            'dropdb', '-U', 'postgres', 'db', '--if-exists',
        ])  # TODO: silence me

        click.echo(' [{}s]'.format(int(time() - start_remove)))

    def get_db_restore_command(self):
        raise NotImplementedError

    def restore_db(self):
        click.secho(' ---> Importing database', nl=False)
        start_import = time()

        db_container_id = utils.get_db_container_id(self.path)

        check_call([
            'docker', 'exec', db_container_id,
            'createdb', '-U', 'postgres', 'db',
        ])

        if self.database_extensions:
            available_extensions = check_output([
                'docker', 'exec', db_container_id,
                'psql', '-U', 'postgres', '--dbname=postgres',
                '-c', 'SELECT name FROM pg_catalog.pg_available_extensions',
            ])

            # TODO: solve extensions in a generic way in
            # harmony with server side db-api
            click.echo('')
            for extension in self.database_extensions:
                if extension in available_extensions:
                    click.echo(
                        '      Enabling extension: {}'.format(extension),
                    )
                    check_call([
                        'docker', 'exec', db_container_id,
                        'psql', '-U', 'postgres', '--dbname=db', '-c',
                        'CREATE EXTENSION IF NOT EXISTS {};'.format(extension),
                    ], silent=True)

        restore_command = self.get_db_restore_command()

        # TODO: use same dump-type detection like server side on db-api
        try:
            subprocess.call((
                'docker', 'exec', db_container_id,
                '/bin/bash', '-c', restore_command
            ))
        except subprocess.CalledProcessError:
            pass

        click.echo('\n      [{}s]'.format(int(time() - start_import)))

    def finish(self):
        click.secho('Done', fg='green', nl=False)
        click.echo(' [{}s]'.format(int(time() - self.start_time)))


class ImportLocalDatabase(DatabaseImportBase):
    def __init__(self, *args, **kwargs):
        self.custom_dump_path = kwargs.pop('custom_dump_path')
        super(ImportLocalDatabase, self).__init__(*args, **kwargs)

    def setup(self):
        click.secho(
            ' ===> Loading database dump {} into local database'
            .format(self.custom_dump_path)
        )
        db_container_id = utils.get_db_container_id(self.path)

        start_copy = time()

        click.secho(' ---> Copying dump into container', nl=False)
        check_call([
            'docker', 'cp', self.custom_dump_path,
            '{}:/tmp/dump'.format(db_container_id),
        ], catch=False, silent=True)
        click.echo(' [{}s]'.format(int(time() - start_copy)))
        self.db_dump_path = '/tmp/dump'

    def get_db_restore_command(self):
        if self.custom_dump_path.endswith('sql'):
            kind = 'sql'
        else:
            kind = 'binary'
        return self.restore_commands[kind].format(self.db_dump_path)


class ImportRemoteDatabase(DatabaseImportBase):
    def __init__(self, *args, **kwargs):
        self.stage = kwargs.pop('stage', None)
        super(ImportRemoteDatabase, self).__init__(*args, **kwargs)
        click.secho(
            ' ===> Pulling database from {} {} server'
            .format(self.website_slug, self.stage)
        )

    def setup(self):
        click.secho(' ---> Preparing download', nl=False)
        start_preparation = time()
        response = self.client.download_db_request(
            self.website_id, self.stage) or {}
        progress_url = response.get('progress_url')
        if not progress_url:
            click.secho(' error!', fg='red')
            sys.exit(1)
        progress = {'success': None}
        while progress.get('success') is None:
            sleep(2)
            progress = self.client.download_db_progress(url=progress_url)
        if not progress.get('success'):
            click.secho(' error!', fg='red')
            click.secho(progress.get('result') or '')
            sys.exit(1)
        download_url = progress.get('result') or None
        click.echo(' [{}s]'.format(int(time() - start_preparation)))

        click.secho(' ---> Downloading database', nl=False)
        start_download = time()
        db_dump_path = download_file(download_url, directory=self.path)
        click.echo(' [{}s]'.format(int(time() - start_download)))
        # strip path from dump_path for use in the docker container
        self.db_dump_path = '/app/{}'.format(db_dump_path.replace(self.path, ''))

    def get_db_restore_command(self):
        cmd = self.restore_commands['archived-binary']
        return cmd.format(self.db_dump_path)


def pull_media(client, stage, path=None):
    project_home = utils.get_project_home(path)
    path = os.path.join(project_home, 'data', 'media')
    website_id = utils.get_aldryn_project_settings(path)['id']
    website_slug = utils.get_aldryn_project_settings(path)['slug']

    click.secho(
        ' ===> Pulling media files from {} {} server'.format(
            website_slug,
            stage,
        ),
    )
    start_time = time()
    click.secho(' ---> Preparing download', nl=False)
    start_preparation = time()
    response = client.download_media_request(website_id, stage) or {}
    progress_url = response.get('progress_url')
    if not progress_url:
        click.secho(' error!', fg='red')
        sys.exit(1)

    progress = {'success': None}
    while progress.get('success') is None:
        sleep(2)
        progress = client.download_media_progress(url=progress_url)
    if not progress.get('success'):
        click.secho(' error!', fg='red')
        click.secho(progress.get('result') or '')
        sys.exit(1)
    download_url = progress.get('result') or None
    click.echo(' [{}s]'.format(int(time() - start_preparation)))

    click.secho(' ---> Downloading', nl=False)
    start_download = time()
    backup_path = download_file(download_url)
    if not backup_path:
        # no backup yet, skipping
        return
    click.echo(' [{}s]'.format(int(time() - start_download)))

    if os.path.isdir(path):
        start_remove = time()
        click.secho(' ---> Removing local files', nl=False)
        shutil.rmtree(path)
        click.echo(' [{}s]'.format(int(time() - start_remove)))

    if 'linux' in sys.platform:
        # On Linux, Docker typically runs as root, so files and folders
        # created from within the container will be owned by root. As a
        # workaround, make the folder permissions more permissive, to
        # allow the invoking user to create files inside it.
        docker_compose = utils.get_docker_compose_cmd(project_home)
        check_call(
            docker_compose(
                'run', '--rm', 'web',
                'chown', '-R', str(os.getuid()), 'data'
            )
        )

    click.secho(' ---> Extracting files to {}'.format(path), nl=False)
    start_extract = time()
    with open(backup_path, 'rb') as fobj:
        with tarfile.open(fileobj=fobj, mode='r:*') as media_archive:
            media_archive.extractall(path=path)
    os.remove(backup_path)
    click.echo(' [{}s]'.format(int(time() - start_extract)))
    click.secho('Done', fg='green', nl=False)
    click.echo(' [{}s]'.format(int(time() - start_time)))


def dump_database(dump_filename, archive_filename=None):
    project_home = utils.get_project_home()
    docker_compose = utils.get_docker_compose_cmd(project_home)
    utils.start_database_server(docker_compose)

    click.secho(' ---> Dumping local database', nl=False)
    start_dump = time()
    db_container_id = utils.get_db_container_id(project_home)
    subprocess.call((
        'docker', 'exec', db_container_id,
        'pg_dump', '-U', 'postgres', '-d', 'db',
        '--no-owner', '--no-privileges',
        '-f', os.path.join('/app/', dump_filename)
    ))
    click.echo(' [{}s]'.format(int(time() - start_dump)))

    if not archive_filename:
        # archive filename not specified
        # return path to uncompressed dump
        return os.path.join(project_home, dump_filename)

    archive_path = os.path.join(project_home, archive_filename)
    sql_dump_size = os.path.getsize(dump_filename)
    click.secho(
        ' ---> Compressing SQL dump ({})'.format(pretty_size(sql_dump_size)),
        nl=False,
    )
    start_compress = time()
    with tarfile.open(archive_path, mode='w:gz') as tar:
        tar.add(
            os.path.join(project_home, dump_filename),
            arcname=dump_filename
        )
    compressed_size = os.path.getsize(archive_filename)
    click.echo(
        ' {} [{}s]'.format(
            pretty_size(compressed_size),
            int(time() - start_compress),
        )
    )


DEFAULT_DUMP_FILENAME = 'local_db.sql'


def export_db():
    dump_filename = DEFAULT_DUMP_FILENAME

    click.secho(' ===> Exporting local database to {}'.format(dump_filename))
    start_time = time()

    dump_database(dump_filename=dump_filename)

    click.secho('Done', fg='green', nl=False)
    click.echo(' [{}s]'.format(int(time() - start_time)))


def push_db(client, stage):
    project_home = utils.get_project_home()
    website_id = utils.get_aldryn_project_settings(project_home)['id']
    dump_filename = DEFAULT_DUMP_FILENAME
    archive_filename = dump_filename.replace('.sql', '.tar.gz')
    website_slug = utils.get_aldryn_project_settings(project_home)['slug']

    click.secho(
        ' ===> Pushing local database to {} {} server'.format(
            website_slug,
            stage,
        ),
    )
    start_time = time()

    dump_database(
        dump_filename=dump_filename,
        archive_filename=archive_filename,
    )

    click.secho(' ---> Uploading', nl=False)
    start_upload = time()
    response = client.upload_db(website_id, stage, archive_path) or {}
    click.echo(' [{}s]'.format(int(time() - start_upload)))

    progress_url = response.get('progress_url')
    if not progress_url:
        click.secho(' error!', fg='red')
        sys.exit(1)

    click.secho(' ---> Processing', nl=False)
    start_processing = time()
    progress = {'success': None}
    while progress.get('success') is None:
        sleep(2)
        progress = client.upload_db_progress(url=progress_url)
    if not progress.get('success'):
        click.secho(' error!', fg='red')
        click.secho(progress.get('result') or '')
        sys.exit(1)
    click.echo(' [{}s]'.format(int(time() - start_processing)))

    # clean up
    for temp_file in (dump_filename, archive_filename):
        os.remove(os.path.join(project_home, temp_file))
    click.secho('Done', fg='green', nl=False)
    click.echo(' [{}s]'.format(int(time() - start_time)))


def push_media(client, stage):
    project_home = utils.get_project_home()
    website_id = utils.get_aldryn_project_settings(project_home)['id']
    archive_path = os.path.join(project_home, 'local_media.tar.gz')
    website_slug = utils.get_aldryn_project_settings(project_home)['slug']
    click.secho(
        ' ---> Pushing local media to {} {} server'.format(
            website_slug,
            stage,
        ),
    )
    start_time = time()
    click.secho('Compressing local media folder',  nl=False)
    uncompressed_size = 0
    start_compression = time()
    with tarfile.open(archive_path, mode='w:gz') as tar:
        media_dir = os.path.join(project_home, 'data', 'media')
        for item in os.listdir(media_dir):
            if item == 'MANIFEST':
                # partial uploads are currently not supported
                # not including MANIFEST to do a full restore
                continue
            file_path = os.path.join(media_dir, item)
            tar.add(file_path, arcname=item)
            uncompressed_size += get_size(file_path)
        file_count = len(tar.getmembers())
    click.echo(
        ' {} {} ({}) compressed to {} [{}s]'.format(
            file_count,
            'files' if file_count > 1 else 'file',
            pretty_size(uncompressed_size),
            pretty_size(os.path.getsize(archive_path)),
            int(time() - start_compression),
        )
    )
    click.secho('Uploading', nl=False)
    start_upload = time()
    response = client.upload_media(website_id, stage, archive_path) or {}
    click.echo(' [{}s]'.format(int(time() - start_upload)))
    progress_url = response.get('progress_url')
    if not progress_url:
        click.secho(' error!', fg='red')
        sys.exit(1)

    click.secho('Processing', nl=False)
    start_processing = time()
    progress = {'success': None}
    while progress.get('success') is None:
        sleep(2)
        progress = client.upload_media_progress(url=progress_url)
    if not progress.get('success'):
        click.secho(' error!', fg='red')
        click.secho(progress.get('result') or '')
        sys.exit(1)
    click.echo(' [{}s]'.format(int(time() - start_processing)))

    # clean up
    os.remove(archive_path)
    click.secho('Done', fg='green', nl=False)
    click.echo(' [{}s]'.format(int(time() - start_time)))


def update_local_project(git_branch):
    project_home = utils.get_project_home()
    docker_compose = utils.get_docker_compose_cmd(project_home)

    click.secho('Pulling changes from git remote', fg='green')
    check_call(('git', 'pull', 'origin', git_branch))
    click.secho('Pulling docker images', fg='green')
    check_call(docker_compose('pull'))
    click.secho('Building local docker images', fg='green')
    check_call(docker_compose('build'))


def develop_package(package, no_rebuild=False):
    """
    :param package: package name in addons-dev folder
    """

    project_home = utils.get_project_home()
    addons_dev_dir = os.path.join(project_home, 'addons-dev')

    if not os.path.isdir(os.path.join(addons_dev_dir, package)):
        raise click.ClickException(
            'Package {} could not be found in {}. Please make '
            'sure it exists and try again.'
            .format(package, addons_dev_dir)
        )

    url_pattern = re.compile('(\S*/{}/\S*)'.format(package))
    new_package_path = '-e /app/addons-dev/{}\n'.format(package)

    # add package to requirements.in for dependencies
    requirements_file = os.path.join(project_home, 'requirements.in')
    # open file with 'universal newline support'
    # https://docs.python.org/2/library/functions.html#open
    with open(requirements_file, 'rU') as fh:
        addons = fh.readlines()

    replaced = False

    for counter, addon in enumerate(addons):
        if re.match(url_pattern, addon) or addon == new_package_path:
            addons[counter] = new_package_path
            replaced = True
            break

    if not replaced:
        # Not replaced, append to generated part of requirements
        for counter, addon in enumerate(addons):
            if '</INSTALLED_ADDONS>' in addon:
                addons.insert(counter, new_package_path)
                replaced = True
                break

    if not replaced:
        # fallback: generated section seems to be missing, appending
        addons.append(new_package_path)

    with open(requirements_file, 'w') as fh:
        fh.writelines(addons)

    if not no_rebuild:
        # build web again
        docker_compose = utils.get_docker_compose_cmd(project_home)

        check_call(docker_compose('build', 'web'))

    click.secho(
        'The package {} has been added to your local development project!'
        .format(package)
    )


def open_project(open_browser=True):
    docker_compose = utils.get_docker_compose_cmd(utils.get_project_home())
    try:
        addr = check_output(docker_compose('port', 'web', '80'), catch=False)
    except subprocess.CalledProcessError:
        if click.prompt('Your project is not running. Do you want to start '
                        'it now?'):
            return start_project()
        return
    host, port = addr.rstrip(os.linesep).split(':')

    if host == '0.0.0.0':
        docker_host_url = os.environ.get('DOCKER_HOST')
        if docker_host_url:
            proto, host_port = os.environ.get('DOCKER_HOST').split('://')
            host = host_port.split(':')[0]

    addr = 'http://{}:{}/'.format(host, port)

    click.secho(
        'Your project is configured to run at {}'.format(addr),
        fg='green'
    )

    click.secho('Waiting for project to start..', fg='green', nl=False)
    # wait 30s for runserver to startup
    seconds = 30
    for attempt in range(seconds):
        click.secho('.', fg='green', nl=False)
        try:
            requests.head(addr)
        except requests.ConnectionError:
            sleep(1)
        else:
            click.echo()
            break
    else:
        raise click.ClickException(
            "\nProject failed to start. Please run 'docker-compose logs' "
            "to get more information."
        )

    if open_browser:
        click.launch(addr)
    return addr


def start_project():
    docker_compose = utils.get_docker_compose_cmd(utils.get_project_home())
    try:
        check_output(docker_compose('up', '-d'), catch=False, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as exc:
        output = exc.output.decode()
        if 'port is already allocated' in output:
            click.secho(
                "There's already another program running on this project's "
                "port. Please either stop the other program or change the "
                "port in the 'docker-compose.yml' file and try again.\n",
                fg='red'
            )
        raise click.ClickException(output)

    return open_project(open_browser=True)


def show_project_status():
    docker_compose = utils.get_docker_compose_cmd(utils.get_project_home())
    check_call(docker_compose('ps'))


def stop_project():
    docker_compose = utils.get_docker_compose_cmd(utils.get_project_home())
    check_call(docker_compose('stop'))



import glob
import os
import tarfile

import click

from .. import settings
from ..utils import get_bytes_io
from ..validators.common import load_config
from ..validators.boilerplate import validate_boilerplate

from .common import add_meta_files


BOILERPLATE_EXCLUDE_DEFAULTS = [
    'boilerplate.json',
    '.git',
]


def normalize_path(path):
    return os.path.normpath(path)


def get_boilerplate_files(boilerplate_path):
    config = load_config(settings.BOILERPLATE_CONFIG_FILENAME, boilerplate_path)
    excluded_patterns = config.get('excluded', []) + BOILERPLATE_EXCLUDE_DEFAULTS
    # glob excludes
    excluded = []
    for exclude in excluded_patterns:
        excluded += glob.glob(normalize_path(exclude).rstrip('/'))

    excluded = set(excluded)
    matches = []

    for path, subdirs, files in os.walk(boilerplate_path, topdown=True):
        subdirs[:] = [
            sub for sub in subdirs
            if normalize_path(os.path.join(path, sub)) not in excluded
        ]

        if normalize_path(path) not in excluded:  # check root level excludes
            for fname in files:
                fpath = os.path.join(path, fname)
                if normalize_path(fpath) not in excluded:
                    matches.append(fpath)

    return excluded, matches


def upload_boilerplate(client, path=None, noinput=False):
    path = path or '.'
    errors = validate_boilerplate(path)

    if errors:
        message = 'The following errors happened during validation:'
        message = '{}\n - {}'.format(message, '\n - '.join(errors))
        raise click.ClickException(message)

    excludes, files = get_boilerplate_files(path)

    if not noinput:
        click.secho(
            'The following files will be included in your '
            'boilerplate and uploaded to Aldryn:'.format(len(files)),
            fg='yellow'
        )
        click.echo(os.linesep.join(files))
        click.confirm(
            'Are you sure you want to continue and upload '
            'the preceding (#{}) files to Aldryn?'.format(len(files)),
            default=True, show_default=True, abort=True
        )

    archive_obj = create_boilerplate_archive(path, files)
    return client.upload_boilerplate(archive_obj)


def create_boilerplate_archive(path, files):
    fobj = get_bytes_io()

    with tarfile.open(mode='w:gz', fileobj=fobj) as tar:
        add_meta_files(tar, path, settings.BOILERPLATE_CONFIG_FILENAME)
        for f in files:
            tar.add(f)

    fobj.seek(0)
    return fobj



import os

import click
import requests
from six.moves.urllib_parse import urljoin

from . import messages
from .utils import create_temp_dir, get_user_agent


class SingleHostSession(requests.Session):
    def __init__(self, host, **kwargs):
        super(SingleHostSession, self).__init__()
        self.debug = kwargs.pop('debug', False)
        self.host = host.rstrip('/')
        for key, value in kwargs.items():
            setattr(self, key, value)

    def request(self, method, url, *args, **kwargs):
        url = urljoin(self.host, url)
        return super(SingleHostSession, self).request(
            method, url, *args, **kwargs
        )


class APIRequest(object):
    network_exception_message = messages.NETWORK_ERROR_MESSAGE
    default_error_message = messages.SERVER_ERROR
    response_code_error_map = {
        requests.codes.forbidden: messages.AUTH_INVALID_TOKEN,
        requests.codes.not_found: messages.RESOURCE_NOT_FOUND,
    }

    method = 'GET'
    url = None
    default_headers = {
        'User-Agent': get_user_agent(),
    }
    headers = {}

    def __init__(self, session, url=None, url_kwargs=None, data=None,
                 files=None, *args, **kwargs):
        self.session = session
        if url:
            self.url = url
        self.url_kwargs = url_kwargs or {}
        self.data = data or {}
        self.files = files or {}

    def __call__(self, *args, **kwargs):
        return self.request(*args, **kwargs)

    def get_url(self):
        return self.url.format(**self.url_kwargs)

    def get_error_code_map(self):
        return self.response_code_error_map

    def get_headers(self):
        headers = self.default_headers.copy()
        headers.update(self.headers)
        return headers

    def request(self, *args, **kwargs):
        try:
            response = self.session.request(
                self.method, self.get_url(),
                data=self.data, files=self.files,
                headers=self.get_headers(),
                *args, **kwargs
            )
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout) as e:
            raise click.ClickException(
                messages.NETWORK_ERROR_MESSAGE + unicode(e)
            )

        return self.verify(response)

    def verify(self, response):
        if not response.ok:
            error_msg = self.get_error_code_map().get(response.status_code)
            if not error_msg:
                response_content = response.content
                if not self.session.debug:
                    response_content = response_content[:300]
                error_msg = '{}\n\n{}'.format(
                    self.default_error_message,
                    response_content,
                )
            raise click.ClickException(error_msg)
        return self.process(response)

    def process(self, response):
        return response.json()


class RawResponse(object):
    def process(self, response):
        return response


class TextResponse(object):
    def process(self, response):
        return response.text


class JsonResponse(object):
    def process(self, response):
        return response.json()


class DjangoFormMixin(object):
    success_message = 'Request successful'

    def verify(self, response):
        if response.ok:
            return self.success_message
        elif response.status_code == requests.codes.bad_request:
            formatted = (
                'There was an error submitting your request:\n'
                '-------------------------------------------\n\n'
            )
            for field, errors in response.json().items():
                formatted += ' - {}\n'.format(field)
                for error in errors:
                    formatted += '   - {}\n'.format(error)
                formatted += '\n'
            return formatted.strip('\n')
        return super(DjangoFormMixin, self).verify(response)


class FileResponse(object):
    def __init__(self, *args, **kwargs):
        self.filename = kwargs.pop('filename', None)
        self.directory = kwargs.pop('directory', None)
        super(FileResponse, self).__init__(*args, **kwargs)

    def process(self, response):
        dump_path = os.path.join(
            self.directory or create_temp_dir(),
            self.filename or 'data.tar.gz',
        )

        with open(dump_path, 'wb') as f:
            for chunk in response.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)
                    f.flush()
        return dump_path

    def request(self, *args, **kwargs):
        kwargs['stream'] = True
        return super(FileResponse, self).request(*args, **kwargs)


class LoginRequest(APIRequest):
    default_error_message = messages.AUTH_SERVER_ERROR
    url = '/api/v1/login-with-token/'
    method = 'POST'


class LoginStatusRequest(APIRequest):
    url = '/track/'
    method = 'GET'


class ProjectListRequest(APIRequest):
    url = '/api/v1/user-websites/'


class ProjectDetailRequest(APIRequest):
    url = '/api/v1/website/{website_id}/detail/'


class DeployProjectProgressRequest(JsonResponse, APIRequest):
    url = '/api/v1/website/{website_id}/deploy/'
    method = 'GET'


class DeployProjectRequest(JsonResponse, APIRequest):
    url = '/api/v1/website/{website_id}/deploy/'
    method = 'POST'


class RegisterAddonRequest(DjangoFormMixin, JsonResponse, APIRequest):
    url = '/api/v1/addon/register/'
    method = 'POST'
    success_message = 'Addon successfully registered'


class UploadAddonRequest(TextResponse, APIRequest):
    url = '/api/v1/apps/'
    method = 'POST'


class UploadBoilerplateRequest(TextResponse, APIRequest):
    url = '/api/v1/boilerplates/'
    method = 'POST'


class ProjectLockQueryRequest(APIRequest):
    url = '/api/v1/website/{website_id}/lock/'
    method = 'GET'

    def process(self, response):
        return response.json('is_locked')


class ProjectLockRequest(TextResponse, APIRequest):
    url = '/api/v1/website/{website_id}/lock/'
    method = 'PUT'


class ProjectUnlockRequest(TextResponse, APIRequest):
    url = '/api/v1/website/{website_id}/lock/'
    method = 'DELETE'


class SlugToIDRequest(APIRequest):
    url = '/api/v1/slug-to-id/{website_slug}/'

    def process(self, response):
        return response.json().get('id')


class IDToSlugRequest(APIRequest):
    url = '/api/v1/id-to-slug/{website_id}/'

    def process(self, response):
        return response.json().get('slug')


class DownloadBackupRequest(FileResponse, APIRequest):
    url = '/api/v1/workspace/{website_slug}/download/backup/'
    headers = {'accept': 'application/x-tar-gz'}

    def verify(self, response):
        if response.status_code == requests.codes.not_found:
            # no backups yet, ignore
            return None
        return super(DownloadBackupRequest, self).verify(response)


# Download DB

class DownloadDBRequestRequest(JsonResponse, APIRequest):
    url = '/api/v1/website/{website_id}/download/db/request/'
    method = 'POST'


class DownloadDBProgressRequest(JsonResponse, APIRequest):
    method = 'GET'


# Download Media

class DownloadMediaRequestRequest(JsonResponse, APIRequest):
    url = '/api/v1/website/{website_id}/download/media/request/'
    method = 'POST'


class DownloadMediaProgressRequest(JsonResponse, APIRequest):
    method = 'GET'


# Upload DB

class UploadDBRequest(JsonResponse, APIRequest):
    url = '/api/v1/website/{website_id}/upload/db/'
    method = 'POST'

    def get_error_code_map(self):
        error_codes = super(UploadDBRequest, self).get_error_code_map()
        error_codes[requests.codes.bad_request] = messages.INVALID_DB_SUBMITTED
        return error_codes

    def verify(self, response):
        if response.status_code == requests.codes.bad_request:
            try:
                db_log = response.json()['message'].encode('utf-8')
            except (TypeError, IndexError):
                pass
            else:
                logfile = os.path.join(os.getcwd(), 'db_upload.log')
                with open(logfile, 'w+') as fh:
                    fh.write(db_log)

        return super(UploadDBRequest, self).verify(response)


class UploadDBProgressRequest(JsonResponse, APIRequest):
    method = 'GET'


# Upload Media

class UploadMediaFilesRequest(JsonResponse, APIRequest):
    url = '/api/v1/website/{website_id}/upload/media/'
    method = 'POST'


class UploadMediaFilesProgressRequest(JsonResponse, APIRequest):
    method = 'GET'



import subprocess
import platform
import tarfile
import tempfile
import io
import os
import sys
from subprocess import CalledProcessError
from contextlib import contextmanager
from distutils.version import StrictVersion
from math import log

import click
import requests
from six import PY2
from tabulate import tabulate
from six.moves.urllib_parse import urljoin

from . import __version__


ALDRYN_DEFAULT_BRANCH_NAME = "develop"


def hr(char='-', width=None, **kwargs):
    if width is None:
        width = click.get_terminal_size()[0]
    click.secho(char * width, **kwargs)


def table(data, headers):
    return tabulate(data, headers)


def indent(text, spaces=4):
    return '\n'.join(' ' * spaces + ln for ln in text.splitlines())


def get_package_version(path):
    return check_output(
            ['python', 'setup.py', '--version'],
            cwd=path
    ).strip()


@contextmanager
def dev_null():
    with open(os.devnull, 'wb') as devnull:
        yield devnull


@contextmanager
def silence_stderr():
    with dev_null() as devnull:
        with redirect_stderr(devnull):
            yield


@contextmanager
def silence_stdout():
    with dev_null() as devnull:
        with redirect_stdout(devnull):
            yield


@contextmanager
def redirect_stdout(new_stream):
    original_stream = sys.stdout
    sys.stdout = new_stream
    try:
        yield
    finally:
        sys.stdout = original_stream


@contextmanager
def redirect_stderr(new_stream):
    original_stream = sys.stderr
    sys.stderr = new_stream
    try:
        yield
    finally:
        sys.stderr = original_stream


def create_temp_dir():
    return tempfile.mkdtemp(prefix='tmp_aldryn_client_')


def get_bytes_io(*args, **kwargs):
    if PY2:
        from StringIO import StringIO
        cls = StringIO
    else:
        from io import BytesIO
        cls = BytesIO
    return cls(*args, **kwargs)


def get_string_io(*args, **kwargs):
    if PY2:
        from StringIO import StringIO
    else:
        from io import StringIO
    return StringIO(*args, **kwargs)


def tar_add_stringio(tar, string_io, name):
    bytes_io = io.BytesIO(string_io.getvalue().encode())
    return tar_add_bytesio(tar, bytes_io, name)


def tar_add_bytesio(tar, bytes_io, name):
    info = tarfile.TarInfo(name=name)
    bytes_io.seek(0, os.SEEK_END)
    info.size = bytes_io.tell()
    bytes_io.seek(0)
    tar.addfile(tarinfo=info, fileobj=bytes_io)


def execute(func, *popenargs, **kwargs):
    catch = kwargs.pop('catch', True)
    if kwargs.pop('silent', False):
        if 'stdout' not in kwargs:
            kwargs['stdout'] = open(os.devnull, 'w')
            if not is_windows():
                # close file descriptor devnull after exit
                # unfortunately, close_fds is not supported on Windows
                # platforms if you redirect stdin/stdout/stderr
                # => http://svn.python.org/projects/python/
                #    branches/py3k/Lib/subprocess.py
                kwargs['close_fds'] = True
        if 'stderr' not in kwargs:
            kwargs['stderr'] = subprocess.STDOUT
    try:
        return func(*popenargs, **kwargs)
    except subprocess.CalledProcessError as exc:
        if not catch:
            raise
        output = (
            'There was an error trying to run a command. This is most likely',
            'not an issue with aldryn-client, but the called program itself.',
            'Try checking the output of the command above.',
            'The command was:',
            '  {command}'.format(command=' '.join(exc.cmd))
        )
        hr(fg='red')
        click.secho(os.linesep.join(output), fg='red')
        sys.exit(1)


def check_call(*popenargs, **kwargs):
    return execute(subprocess.check_call, *popenargs, **kwargs)


def check_output(*popenargs, **kwargs):
    return execute(subprocess.check_output, *popenargs, **kwargs).decode()


def open_project_cloud_site(client, stage):
    from .localdev.utils import get_aldryn_project_settings

    assert stage in ('test', 'live')
    project_settings = get_aldryn_project_settings()
    project_data = client.get_project(project_settings['id'])
    url = project_data['{}_status'.format(stage)]['site_url']
    if url:
        click.launch(url)
    else:
        click.secho('No {} server deployed yet.'.format(stage), fg='yellow')


def get_dashboard_url(client):
    from .localdev.utils import get_aldryn_project_settings

    project_settings = get_aldryn_project_settings()
    project_data = client.get_project(project_settings['id'])
    return project_data['dashboard_url']


def get_project_cheatsheet_url(client):
    dashboard = get_dashboard_url(client)
    return urljoin(dashboard, 'local-development/')


def is_windows():
    return sys.platform == 'win32'


def is_linux():
    return sys.platform.startswith('linux')


unit_list = list(zip(
        ['bytes', 'kB', 'MB', 'GB', 'TB', 'PB'],
        [0, 0, 1, 2, 2, 2],
))


def pretty_size(num):
    """Human friendly file size"""
    # http://stackoverflow.com/a/10171475
    if num > 1:
        exponent = min(int(log(num, 1024)), len(unit_list) - 1)
        quotient = float(num) / 1024 ** exponent
        unit, num_decimals = unit_list[exponent]
        format_string = '{:.%sf} {}' % (num_decimals)
        return format_string.format(quotient, unit)
    elif num == 0:
        return '0 bytes'
    elif num == 1:
        return '1 byte'


def get_size(start_path):
    """
    Get size of the file or directory specified by start_path in bytes.

    If ``start_path`` points to a file - get it's size, if it points to a
    directory - calculate total size of all the files within it
    (including subdirectories).
    """
    # http://stackoverflow.com/a/1392549/176490

    if os.path.isfile(start_path):
        return os.path.getsize(start_path)

    total_size = 0
    for dirpath, dirnames, filenames in os.walk(start_path):
        for filename in filenames:
            fp = os.path.join(dirpath, filename)
            total_size += os.path.getsize(fp)
    return total_size


def get_latest_version_from_pypi():
    try:
        response = requests.get(
            'https://pypi.python.org/pypi/aldryn-client/json'
        )
        response.raise_for_status()
        newest_version = StrictVersion(response.json()['info']['version'])
        return newest_version, None
    except requests.RequestException as exc:
        return False, exc
    except (KeyError, ValueError):
        return False, None


def get_git_commit():
    script_home = os.path.dirname(__file__)
    git_dir = os.path.join(script_home, '..', '.git')
    if os.path.exists(git_dir):
        try:
            return subprocess.check_output([
                'git', '--git-dir', git_dir,
                'rev-parse', '--short', 'HEAD'
            ]).strip()
        except:
            pass


def get_git_checked_branch():
    try:
        return subprocess.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD']).strip()
    except CalledProcessError:
        return ALDRYN_DEFAULT_BRANCH_NAME


def get_user_agent():
    revision = get_git_commit()
    if revision:
        client = 'aldryn-client/{}-{}'.format(__version__, revision)
    else:
        client = 'aldryn-client/{}'.format(__version__)

    os = '{}/{}'.format(platform.system(), platform.release())
    python = '{}/{}'.format(
        platform.python_implementation(),
        platform.python_version(),
    )
    return '{} ({}; {})'.format(client, os, python)


def download_file(url, directory=None, filename=None):
    response = requests.get(url, stream=True)

    dump_path = os.path.join(
        directory or create_temp_dir(),
        filename or 'data.tar.gz',
    )

    with open(dump_path, 'wb') as f:
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:  # filter out keep-alive new chunks
                f.write(chunk)
                f.flush()
    return dump_path
